require( "internal.lua" )

class( "Color" )

function Color._parse_args( args )
	local _l = #args
	if _l == 0 then
		return { 0.0, 0.0, 0.0, 1.0 }
	elseif _l == 1 then
		local a = args[1]
		if __type( a ) == "Color" then
			return { a.r, a.g, a.b, a.a }
		elseif __type( a ) == "table" then
			return { a[1], a[2] or 0.0, a[3] or 0.0, a[4] or 1.0 }
		elseif __type( a ) == "string" then
			local c = Color.FromString( a )
			return { c.r, c.g, c.b, c.a }
		end
		error( "Cannot initialize Color; Check arguments: "..__type( a ) )
	elseif _l >= 2 then
		return { args[1], args[2], args[3] or 0.0, args[4] or 0.0 }
	end
	error( "Cannot initialize Color; Check arguments: "..__type( args ) )
end

function Color:__init( ... )
	local args = Color._parse_args( table.pack( ... ) )
	self.r = args[1]
	self.g = args[2]
	self.b = args[3]
	self.a = args[4]
end

function Color:__add( other )
	return Color( self.r + other.r, self.g + other.g, self.b + other.b, self.a )
end

function Color:__sub( other )
	return Color( self.r - other.r, self.g - other.g, self.b - other.b, self.a )
end

function Color:__mul( other )
	local _type = __type( other )
	if _type == "number" then
		return Color( self.r * other , self.g * other , self.b * other, self.a )
	end
	if _type == "Color" then
		return Color( self.r * other.r , self.g * other.g , self.b * other.b, self.a )
	end
	error( "Color:__mul Invalid other type: ".._type.." Color or number is required" )
end

function Color:__eq( other )
	return self.r == other.r and self.g == other.g and self.b == other.b and self.a == other.a
end

function Color:__tostring( )
	return string.format( "Color( %.2f, %.2f, %.2f, %.2f )", self.r, self.g, self.b, self.a )
end

function Color:Lerp( other, factor )
	local r = other.r * factor + self.r * ( 1.0 - factor )
	local g = other.g * factor + self.g * ( 1.0 - factor )
	local b = other.b * factor + self.b * ( 1.0 - factor )
	local a = other.a * factor + self.a * ( 1.0 - factor )
	return Color( r, g, b, a )
end

function Color:Luminance()
	return 0.2126 * self.r + 0.7152 * self.g + 0.0722 * self.b
end

function Color:ToInts()
	return math.floor( self.r * 255 ), math.floor( self.g * 255 ), math.floor( self.b * 255 ), math.floor( self.a * 255 )
end

function Color:ToHex()
	return string.format( "#%02x%2x%02x%02x", self:ToInts() )
end

--
-- STATIC METHODS
--

function Color.bezier3( v1, v2, v3, t )
	local v = Color()
	v.r = b.bezier3( v1.r, v2.r, v3.r, t )
	v.g = b.bezier3( v1.g, v2.g, v3.g, t )
	v.b = b.bezier3( v1.b, v2.b, v3.b, t )
	v.a = b.bezier3( v1.a, v2.a, v3.a, t )
	return v
end

function Color.bezier4( v1, v2, v3, v4, t )
	local v = Color()
	v.r = bezier4( v1.r, v2.r, v3.r, v4.r, t )
	v.g = bezier4( v1.g, v2.g, v3.g, v4.g, t )
	v.b = bezier4( v1.b, v2.b, v3.b, v4.b, t )
	v.a = bezier4( v1.a, v2.a, v3.a, v4.a, t )
	return v
end

Color.White = Color( 1, 1, 1, 1 )
Color.Black = Color( 0, 0, 0, 1 )
Color.Red = Color( 1, 0, 0, 1 )
Color.Green = Color( 0, 1, 0, 1 )
Color.Blue = Color( 0, 0, 1, 1 )
Color.Yellow = Color( 1, 1, 0, 1 )
Color.Purple = Color( 1, 0, 1, 1 )
Color.Cyan = Color( 0, 1, 1, 1 )

function Color.FromString( s )
	s = string.lower(s)
	s = string.gsub( s, "#", "" )
	local l = string.len( s )
	print( s, l )
	if l == 2 then
		local v = tonumber( s, 16 ) / 255
		return Color( v, v, v, 1 )
	elseif l == 6 then
		local r = tonumber( string.sub( s, 1, 2 ), 16 ) / 255
		local g = tonumber( string.sub( s, 3, 4 ), 16 ) / 255
		local b = tonumber( string.sub( s, 5, 6 ), 16 ) / 255
		return Color( r, g, b, 1 )
	elseif l == 8 then
		local r = tonumber( string.sub( s, 1, 2 ), 16 ) / 255
		local g = tonumber( string.sub( s, 3, 4 ), 16 ) / 255
		local b = tonumber( string.sub( s, 5, 6 ), 16 ) / 255
		local a = tonumber( string.sub( s, 7, 8 ), 16 ) / 255
		return Color( r, g, b, a )
	end
end
