require( "internal.lua" )

class( "Matrix2" )

function Matrix2._parse_args( args )
	local _l = #args
	if _l == 0 then
		return { 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0 }
	elseif _l == 1 then
		local a = args[1]
		if __type( a ) == "Matrix2" then
			return { a.m00, a.m01, a.m02, a.m10, a.m11, a.m12, a.m20, a.m21, a.m22 }
		elseif __type( a ) == "table" then
			return Matrix:_parse_args( a )
		end
		error( "Cannot Initialize Matrix2; Check arguments: "..type( a ) )
	elseif _l == 6 then
		local a = args
		return { a[1], a[2], a[3], a[4], a[5], a[6], 0.0, 0.0, 1.0 }
	elseif _l == 9 then
		local a = args
		return { a[1], a[2], a[3], a[4], a[5], a[6], a[7], a[8], a[9] }
	end
	error( "Cannot Initialize Matrix2; Check arguments: "..type( args ) )
end

function Matrix2:__init( ... )
	local v = Matrix2._parse_args( table.pack( ... ) )
	self.m00, self.m01, self.m02 = v[1], v[2], v[3]
	self.m10, self.m11, self.m12 = v[4], v[5], v[6]
	self.m20, self.m21, self.m22 = v[7], v[8], v[9]
end

function Matrix2:__tostring( )
	return string.format( "Matrix2( %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f, %.2f )",
								self.m00, self.m01, self.m02,
								self.m10, self.m11, self.m12,
								self.m20, self.m21, self.m22 )
end

function Matrix2:__mul( o )
	return Matrix2(	self.m00 * o.m00 + self.m01 * o.m10 + self.m02 * o.m20 ,
					self.m00 * o.m01 + self.m01 * o.m11 + self.m02 * o.m21 ,
					self.m00 * o.m02 + self.m01 * o.m12 + self.m02 * o.m22 ,

					self.m10 * o.m00 + self.m11 * o.m10 + self.m12 * o.m20 ,
					self.m10 * o.m01 + self.m11 * o.m11 + self.m12 * o.m21 ,
					self.m10 * o.m02 + self.m11 * o.m12 + self.m12 * o.m22 ,

					self.m20 * o.m00 + self.m21 * o.m10 + self.m22 * o.m20 ,
					self.m20 * o.m01 + self.m21 * o.m11 + self.m22 * o.m21 ,
					self.m20 * o.m02 + self.m21 * o.m12 + self.m22 * o.m22 )
end

function Matrix2:_set( v )
	self.m00, self.m01, self.m02 = v[0], v[1], v[2]
	self.m10, self.m11, self.m12 = v[3], v[4], v[5]
	self.m20, self.m21, self.m22 = v[6], v[7], v[8]
end

function Matrix2:SetIdentity()
	self.m00, self.m01, self.m02 = 1.0, 0.0, 0.0
	self.m10, self.m11, self.m12 = 0.0, 1.0, 0.0
	self.m20, self.m21, self.m22 = 0.0, 0.0, 1.0
end

function Matrix2:SetTranslation( t )
	self.m00, self.m01, self.m02 = 1.0, 0.0, 0.0
	self.m10, self.m11, self.m12 = 0.0, 1.0, 0.0
	self.m20, self.m21, self.m22 = t.x, t.y, 1.0
end

function Matrix2:SetScaling( s )
	self.m00, self.m01, self.m02 = s.x, 0.0, 0.0
	self.m10, self.m11, self.m12 = 0.0, s.y, 0.0
	self.m20, self.m21, self.m22 = 0.0, 0.0, 1.0
end

function Matrix2:SetRotation( angle )
	local cs = math.cos( angle )
	local sn = math.sin( angle )
	self.m00, self.m01, self.m02 =  cs,  sn, 0.0
	self.m10, self.m11, self.m12 = -sn,  cs, 0.0
	self.m20, self.m21, self.m22 = 0.0, 0.0, 1.0
end

function Matrix2:Determinant()
	return 	self.m00 * self.m11 * self.m22 + self.m01 * self.m12 * self.m20 +
			self.m02 * self.m10 * self.m21 - self.m02 * self.m11 * self.m20 -
			self.m01 * self.m10 * self.m22 - self.m00 * self.m12 * self.m21
end

function Matrix2:GetTranslation()
	return Vector2( self.m20, self.m21 )
end

function Matrix2:GetScaling()
	return Vector2( math.sqrt( self.m00 * self.m00 + self.m01 * self.m01 ),
					math.sqrt( self.m10 * self.m10 + self.m11 * self.m11 ) )
end

function Matrix2:GetRotation()
	local v = self:TransformVector2x2( Vector2( 1, 0 ) ):Normalized()
	return math.atan2( v.y, v.x )
end

function Matrix2:Decompose()
	return self:GetTranslation(), self:GetRotation(), self:GetScaling()
end

function Matrix2:Inverse()
	local det = self:GetDeterminant()
	if abs( det ) > 0.00001 then
		det = 1.0 / det
	else
		return None
	end

	self.m00 = det * ( self.m11 * self.m22 - self.m21 * self.m12 )
	self.m01 = det * ( self.m02 * self.m21 - self.m22 * self.m01 )
	self.m02 = det * ( self.m01 * self.m12 - self.m11 * self.m02 )

	self.m10 = det * ( self.m12 * self.m20 - self.m22 * self.m10 )
	self.m11 = det * ( self.m00 * self.m22 - self.m20 * self.m02 )
	self.m12 = det * ( self.m02 * self.m10 - self.m12 * self.m00 )

	self.m20 = det * ( self.m10 * self.m21 - self.m20 * self.m11 )
	self.m21 = det * ( self.m01 * self.m20 - self.m21 * self.m00 )
	self.m22 = det * ( self.m00 * self.m11 - self.m10 * self.m01 )
end

function Matrix2:Inversed()
	local det = self:GetDeterminant()
	if abs( det ) > 0.00001 then
		det = 1.0 / det
	else
		return None
	end
	inv = Matrix2()
	inv.m00 = det * ( self.m11 * self.m22 - self.m21 * self.m12 )
	inv.m01 = det * ( self.m02 * self.m21 - self.m22 * self.m01 )
	inv.m02 = det * ( self.m01 * self.m12 - self.m11 * self.m02 )

	inv.m10 = det * ( self.m12 * self.m20 - self.m22 * self.m10 )
	inv.m11 = det * ( self.m00 * self.m22 - self.m20 * self.m02 )
	inv.m12 = det * ( self.m02 * self.m10 - self.m12 * self.m00 )

	inv.m20 = det * ( self.m10 * self.m21 - self.m20 * self.m11 )
	inv.m21 = det * ( self.m01 * self.m20 - self.m21 * self.m00 )
	inv.m22 = det * ( self.m00 * self.m11 - self.m10 * self.m01 )
	return inv
end

function Matrix2:TransformVector( v )
	return Vector2( v.x * self.m00 + v.y * self.m10 + self.m20 , v.x * self.m01 + v.y * self.m11 + self.m21  )
end

function Matrix2:TransformVector2x2( v )
	return Vector2( v.x * self.m00 + v.y * self.m10, v.x * self.m01 + v.y * self.m11  )
end

--
-- STATIC METHODS
--

function Matrix2.Translation( t )
	local m = Matrix2()
	m:SetTranslation( t )
	return m
end

function Matrix2.Scaling( s )
	local m = Matrix2()
	m:SetScaling( s )
	return m
end

function Matrix2.Rotation( angle )
	local m = Matrix2()
	m:SetRotation( angle )
	return m
end
