require( "internal.lua" )

class( "Plane" )

function Plane._parse_args( args )
	local _l = #args
	if _l == 0 then
		return { 0.0, 0.0, 0.0, 0.0 }
	elseif _l == 1 and __type( args[1] ) == "Plane" then
			return { args[1].a, args[1].b, args[1].c, args[1].d }
	elseif _l == 2 then -- POINT AND NORMAL as Vectors
		if __type( args[1] ) == "Vector" and __type( args[2] ) == "Vector" then
			return { args[2].x, args[2].y, args[2].z, -args[1]:Dot( args[2] ) }
		end
	elseif _l == 3 then -- Three points as Vectors
			local p1 = args[1]
			local p2 = args[2]
			local p3 = args[3]
			if __type( p1 ) == "Vector" and __type( p2 ) == "Vector"  and __type( p3 ) == "Vector" then
				local normal = ( p2 - p1 ):Cross( p3 - p1 ):Normalized()
				return { normal.x, normal.y, normal.z, -p1:Dot( normal ) }
			end
	elseif _l == 4 then
		return { args[1], args[2], args[3], args[4] }
	end
	error( "Cannot Initialize Plane; Check arguments: "..type( args ) )
end

function Plane:__init( ... )
	local args = Plane._parse_args( table.pack( ... ) )
	self.a = args[1]
	self.b = args[2]
	self.c = args[3]
	self.d = args[4]
end

function Plane:__tostring( )
	return string.format( "Plane( a=%.2f, b=%.2f, c=%.2f, distance=%.2f )", self.a, self.b, self.c, self.d )
end

function Plane.SetValues( _a, _b, _c, _d )
	self.a = _a
	self.b = _b
	self.c = _c
	self.d = _d
end

function Plane:SetPointNormal( point, normal )
	self.a = normal.x
	self.b = normal.y
	self.c = normal.z
	self.d = -point:Dot( normal )
end

function Plane:SetPoints( p1, p2, p3 )
	self:SetPointNormal( p1, ( p2 - p1 ):Cross( p3 - p1 ):Normalized() )
end

function Plane:Normalize( )
	local len = math.sqrt( self.a * self.a + self.b * self.b + self.c * self.c )
	if len > 0 then
		self:SetValues( self.a / len , self.b / len , self.c / len , self.d / len )
	else
		self:SetValues( 0.0, 0.0, 0.0, 0.0 )
	end
end

function Plane:Normalized( )
	local temp = Plane( self.a, self.b, self.c, self.d )
	temp:Normalize()
	return temp
end

function Plane:Distance( point )
	return self:Dot( point ) + self.d
end

function Plane:Dot( point )
	return Vector( self.a, self.b, self.c ):Dot( point )
end

function Plane:Project( point )
	return point - Vector( self.a, self.b, self.c ) * self:Distance( point )
end

function Plane:GetIntersection( ray )
	if ray:Length() <= 0.0 then return -1.0 end

	local v_dot_n = self:Dot( ray:Direction() )
	if math.abs( v_dot_n ) < 0.000001 then return -1.0 end

	local t = - ( self.d + self:Dot(ray:Start() ) ) / v_dot_n
	if ( t < 0 or t > ray:Length() ) then return -1.0 end

	return t / ray:Length()
end
