require( "internal.lua" )

class( "Quaternion" )

function Quaternion._parse_args( args )
	local _l = #args
	if _l == 0 then
		return { 0.0, 0.0, 0.0, 1.0 }
	elseif _l == 1 then
		local a = args[1]
		if __type( a ) == "Quaternion" then
			return { a.x, a.y, a.z, a.w }
		elseif __type( a ) == "table" then
			return { a[1], a[2] or 0.0, a[3] or 0.0, a[4] or 1.0 }
		end
		error( "Cannot Initialize Quaternion; Check arguments: "..type( a ) )
	elseif _l >= 2 then
		return { args[1], args[2], args[3] or 0.0, args[4] or 1.0 }
	end
	error( "Cannot Initialize Quaternion; Check arguments: "..type( args ) )
end

function Quaternion:__init( ... )
	local args = Quaternion._parse_args( table.pack( ... ) )
	self.x = args[1]
	self.y = args[2]
	self.z = args[3]
	self.w = args[4]
end

function Quaternion:__tostring( )
	return string.format( "Quaternion( %f, %f, %f, %f )", self.x, self.y, self.z, self.w )
end

function Quaternion:__eq( other )
	return self.x ~= other.x or self.y ~= other.y or self.z ~= other.w or self.z ~= other.w
end

function Quaternion:__add( other )
	local result = Quaternion()
	result.x = self.x + other.x
	result.y = self.y + other.y
	result.z = self.z + other.z
	result.w = self.w + other.w
	return result
end

function Quaternion:__mul( other )
	local _type = __type( other )
	if _type == "Quaternion" then
		return Quaternion( other.w * self.x + other.x * self.w + other.y * self.z - other.z * self.y ,
							other.w * self.y + other.y * self.w + other.z * self.x - other.x * self.z ,
							other.w * self.z + other.z * self.w + other.x * self.y - other.y * self.x ,
							other.w * self.w - other.x * self.x - other.y * self.y - other.z * self.z )
	elseif _type == "number" then
		return Quaternion( self.x * other , self.y * other , self.z * other , self.w * other )
	end
	error( "Quaternion.__mul Invalid type: ".._type..". A Quaternion or a number is required" )
end

function Quaternion:SetIdentity()
	self.x, self.y, self.z, self.w = 0.0, 0.0, 0.0, 1.0
end

function Quaternion:LengthSQ()
	return self.x * self.x + self.y * self.y + self.z * self.z + self.w * self.w
end

function Quaternion:Length()
	return math.sqrt( self:LengthSQ() )
end

function Quaternion:Inversed()
	local mag = self.LengthSQ()
	if mag > 0.001 then
		return Quaternion( -self.x / mag, -self.y / mag, -self.z / mag, self.w / mag )
	end
	return Quaternion()
end

function Quaternion:Inverse()
	local inv = self.Inversed()
	self.x = inv.x
	self.y = inv.y
	self.z = inv.z
	self.w = inv.w
end

function Quaternion:Conjugate()
	return Quaternion( -self.x, -self.y, -self.z, self.w )
end

function Quaternion:RotateVector( v )
	local q = Quaternion( self:Conjugate( ) * Quaternion( v.x, v.y, v.z, 1.0 ) ) * self
	return Vector( q.x, q.y, q.z )
end

function Quaternion:Normalized()
	local mag = self:Length()
	if mag > 0.001 then
		return Quaternion( self.x / mag, self.y / mag, self.z / mag, self.w / mag )
	end
	return Quaternion()
end

function Quaternion:Normalize()
	local mag = self:Length()
	if mag > 0.001 then
		self.x = self.x / mag
		self.y = self.y / mag
		self.z = self.z / mag
		self.w = self.w / mag
	else
		self.SetIdentity()
	end
end

function Quaternion:EulerNormalized()
	local alt = self.x * self.y + self.z * self.w
	if alt > 0.499 then
		return Vector( 2.0 * math.atan( self.x, self.w ), math.pi * 0.5, 0.0 )
	end
	if alt < -0.499 then
		return Vector( -2.0 * math.atan( self.x, self.w ), -math.pi * 0.5, 0.0 )
	end
	return Vector( math.atan( 2.0 * ( self.x * self.w - self.y * self.z ), 1.0 - 2.0 * ( self.x * self.x + self.z * self.z ) ),
				   math.atan( 2.0 * ( self.y * self.w - self.x * self.z ), 1.0 - 2.0 * ( self.y * self.y + self.z * self.z ) ),
				   math.asin( 2.0 * alt ) )
end

function Quaternion:Euler()
	local xx = self.x * self.x
	local yy = self.y * self.y
	local zz = self.z * self.z
	local ww = self.w * self.w
	local unit = xx + yy + zz + ww
	local alt = self.x * self.y + self.z * self.w
	if alt > 0.499 * unit then
		return Vector( 2.0 * math.atan( self.x , self.w ), math.pi * 0.5, 0.0 )
	end
	if alt < -0.499 * unit then
		return Vector( -2.0 * math.atan( self.x, self.w ), -math.pi * 0.5, 0.0 )
	end
	return Vector( math.atan( 2.0 * ( self.x * self.w - self.y * self.z ), -xx + yy - zz + ww ),
				   math.atan( 2.0 * ( self.y * self.w - self.x * self.z ), xx - yy - zz + ww ),
				   math.asin( 2.0 * alt / unit ) )
end

function Quaternion:Lerp( other, cf )
	local temp = Quaternion( other )
	local cos_theta = self.x * other.x + self.y * other.y + self.z * other.z + self.w * other.w
	if cos_theta < 0.0 then
		cos_theta = -cos_theta
		temp = Quaternion( -temp.x, -temp.y, -temp.z, -temp.w )
	end
	if cos_theta > 0.95 then
		return Quaternion( self * ( 1.0 - cf ) + temp * cf )
	else
		local theta = math.acos( cos_theta )
		return ( self * math.sin( theta * ( 1.0 - cf ) ) + temp * math.sin( theta * cf ) ) * ( 1.0 / math.sin( theta ) )
	end
end

--
-- STATIC METHODS
--

function Quaternion.OrientX( v )
	local euler = Vector( 0.0, math.deg( math.atan( v.z, v.x ) ), math.deg( math.asin( v.y ) ) )
	return Quaternion.Euler( euler, 'XYZ' )
end

function Quaternion.OrientY( v )
	local euler = Vector( math.deg( math.acos( v.y ) ), math.deg( math.atan( v.x, v.z ) ), 0.0 )
	return Quaternion.Euler( euler, 'XYZ' )
end

function Quaternion.OrientZ( v )
	local euler = Vector( math.deg( -math.asin( v.y ) ), math.deg( math.atan( v.x, v.z ) ) , 0.0 )
	return Quaternion.Euler( euler , 'XYZ' )
end

function Quaternion.RotationX( angle )
	angle = angle * 0.5
	local sin_a = math.sin( angle )
	local cos_a = math.cos( angle )
	return Quaternion( sin_a, 0.0, 0.0, cos_a )
end

function Quaternion.RotationY( angle )
	angle = angle * 0.5
	local sin_a = math.sin( angle )
	local cos_a = math.cos( angle )
	return Quaternion( 0.0, sin_a, 0.0, cos_a )
end

function Quaternion.RotationZ( angle )
	angle = angle * 0.5
	local sin_a = math.sin( angle )
	local cos_a = math.cos( angle )
	return Quaternion( 0.0, 0.0, sin_a, cos_a )
end

function Quaternion.Euler( euler, order )
	local quat_x = Quaternion.RotationX( math.rad( euler.x ) )
	local quat_y = Quaternion.RotationY( math.rad( euler.y ) )
	local quat_z = Quaternion.RotationZ( math.rad( euler.z ) )
	if order == 'XYZ' then
		return Quaternion( ( quat_x * quat_y ) * quat_z )
	elseif order == 'XZY' then
		return Quaternion( ( quat_x * quat_z ) * quat_y )
	elseif order == 'YXZ' then
		return Quaternion( ( quat_y * quat_x ) * quat_z )
	elseif order == 'YZX' then
		return Quaternion( ( quat_y * quat_z ) * quat_x )
	elseif order == 'ZXY' then
		return Quaternion( ( quat_z * quat_x ) * quat_y )
	elseif order == 'ZYX' then
		return Quaternion( ( quat_z * quat_y ) * quat_x )
	end
end

function Quaternion.AxisAngle( axis , angle )
	local _type = __type( axis )
	if _type == "Vector" then
		angle = angle * 0.5
		local sin_a = math.sin( angle )
		local cos_a = math.cos( angle )
		return Quaternion( axis.x * sin_a, axis.y * sin_a, axis.z * sin_a, cos_a )
	end
	error( "Quaternion.AxisAngle Invalid type: ".._type..". Vector is required" )
end
