require( "internal.lua" )

class( "Vector" )

function Vector._parse_args( args )
	local _l = #args
	if _l == 0 then
		return { 0.0, 0.0, 0.0, 1.0 }
	elseif _l == 1 then
		local a = args[1]
		if __type( a ) == "Vector" then
			return { a.x, a.y, a.z, a.w }
		elseif __type( a ) == "table" then
			return { a[1], a[2] or 0.0, a[3] or 0.0, a[4] or 1.0 }
		end
		error( "Cannot Initialize Vector; Check arguments: "..type( a ) )
	elseif _l >= 2 then
		return { args[1], args[2], args[3] or 0.0, args[4] or 1.0 }
	end
	error( "Cannot Initialize Vector; Check arguments: "..type( args ) )
end

function Vector:__init( ... )
	local args = Vector._parse_args( table.pack( ... ) )
	self.x = args[1]
	self.y = args[2]
	self.z = args[3]
	self.w = args[4]
end

function Vector:__add( other )
	return Vector( self.x + other.x , self.y + other.y , self.z + other.z )
end

function Vector:__sub( other )
	return Vector( self.x - other.x , self.y - other.y , self.z - other.z )
end

function Vector:__mul( other )
	local _type = __type( other )
	if _type == "number" then
		return Vector( self.x * other , self.y * other , self.z * other )
	end
	if _type == "Vector" then
		return Vector( self.x * other.x , self.y * other.y , self.z * other.z )
	end
	error( "Vector:__mul Invalid other type: ".._type.." Vector or number is required" )
end

function Vector:__eq( other )
	return self.x == other.x and self.y == other.y and self.z == other.z
end

function Vector:__tostring( )
	return string.format( "Vector( %.2f, %.2f, %.2f )", self.x, self.y, self.z )
end

function Vector:LengthSQ( )
	return self.x * self.x + self.y * self.y + self.z * self.z
end

function Vector:Length( )
	return math.sqrt( self:LengthSQ( ) )
end

function Vector:Normalized( )
	local l = self:Length()
	return Vector( self.x / l, self.y / l, self.z / l )
end

function Vector:Normalize( )
	local l = self:Length()
	self.x = self.x / l
	self.y = self.y / l
	self.z = self.z / l
end

function Vector:DistanceSQ( other )
	return ( self - other ):LengthSQ()
end

function Vector:Distance( other )
	return math.sqrt( self:DistanceSQ( other ) )
end

function Vector:Dot( other )
	return self.x * other.x + self.y * other.y + self.z * other.z
end

function Vector:Cross( other )
	return Vector( 	self.y * other.z - self.z * other.y,
					self.z * other.x - self.x * other.z,
					self.x * other.y - self.y * other.x )
end

function Vector:Lerp( other, factor )
	local x = other.x * factor + self.x * ( 1.0 - factor )
	local y = other.y * factor + self.y * ( 1.0 - factor )
	local z = other.z * factor + self.z * ( 1.0 - factor )
	return Vector( x, y, z )
end

function Vector:Project( other )
	local on = other:Normalized()
	return on * self:Dot( on )
end

function Vector:Reflect( normal )
	return normal * -2 * self:Dot( normal ) + self
end

function Vector:Angle( other )
	return math.atan( self:Cross( other ):Length(), self:Dot( other ) )
end

--
-- STATIC METHODS
--

Vector.Zero = Vector()
Vector.One = Vector( 1, 1, 1 )
Vector.XPos = Vector( 1, 0, 0 )
Vector.XNeg = Vector( -1, 0, 0 )
Vector.YPos = Vector( 0, 1, 0 )
Vector.YNeg = Vector( 0, -1, 0 )
Vector.ZPos = Vector( 0, 0, 1 )
Vector.ZNeg = Vector( 0, 0, -1 )

function Vector.Bezier3( v1, v2, v3, t )
	return Vector(  bezier3( v1.x, v2.x, v3.x, t ),
					bezier3( v1.y, v2.y, v3.y, t ),
					bezier3( v1.z, v2.z, v3.z, t ),
					bezier3( v1.w, v2.w, v3.w, t ) )
end

function Vector.Bezier4( v1, v2, v3, v4, t )
	return Vector(  bezier4( v1.x, v2.x, v3.x, v4.x, t ),
					bezier4( v1.y, v2.y, v3.y, v4.y, t ),
					bezier4( v1.z, v2.z, v3.z, v4.z, t ),
					bezier4( v1.w, v2.w, v3.w, v4.w, t ) )
end
