require( "internal.lua" )

class( "Vector2" )

function Vector2._parse_args( args )
	local _l = #args
	if _l == 0 then
		return { 0.0, 0.0 }
	elseif _l == 1 then
		local a = args[1]
		if __type( a ) == "Vector2" then
			return { a.x, a.y }
		elseif __type( a ) == "table" then
			return { a[1], a[2] }
		end
		error( "Cannot Initialize Vector2; Check arguments: "..type( a ) )
	elseif _l >= 2 then
		return { args[1], args[2] }
	end
	error( "Cannot Initialize Vector2; Check arguments: "..type( args ) )
end

function Vector2:__init( ... )
	local args = Vector2._parse_args( table.pack( ... ) )
	self.x = args[1]
	self.y = args[2]
end

function Vector2:__add( other )
	return Vector2( self.x + other.x , self.y + other.y )
end

function Vector2:__sub( other )
	return Vector2( self.x - other.x , self.y - other.y )
end

function Vector2:__mul( other )
	local _type = __type( other )
	if _type == "number" then
		return Vector2( self.x * other , self.y * other )
	end
	if _type == "Vector2" then
		return Vector( self.x * other.x , self.y * other.y )
	end
	error( "Vector2:__mul Invalid other type: ".._type.." Vector2 or number is required" )
end

function Vector2:__eq( other )
	return self.x == other.x and self.y == other.y
end

function Vector2:__tostring( )
	return string.format( "Vector2( %.2f, %.2f )", self.x, self.y )
end

function Vector2:LengthSQ( )
	return self.x * self.x + self.y * self.y
end

function Vector2:Length( )
	return math.sqrt( self:LengthSQ( ) )
end

function Vector2:Normalized( )
	local l = self:Length()
	return Vector2( self.x / l, self.y / l )
end

function Vector2:Normalize( )
	local l = self:Length()
	self.x = self.x / l
	self.y = self.y / l
end

function Vector2:DistanceSQ( other )
	return ( self - other ):LengthSQ()
end

function Vector2:Distance( other )
	return math.sqrt( self:DistanceSQ( other ) )
end

function Vector2:Dot( other )
	return self.x * other.x + self.y * other.y
end

function Vector2:Cross( other )
	return self.x * other.y - self.y * other.x
end

function Vector2:Lerp( other, factor )
	local x = other.x * factor + self.x * ( 1.0 - factor )
	local y = other.y * factor + self.y * ( 1.0 - factor )
	return Vector2( x, y)
end

function Vector2:Project( other )
	local on = other:Normalized()
	return on * self:Dot( on )
end

function Vector2:Rotate( angle )
	local x = self.x * math.cos( angle ) - self.y * math.sin( angle )
	local y = self.x * math.sin( angle ) + self.y * math.cos( angle )
	return Vector2( x, y )
end

function Vector2:Reflect( normal )
	return normal * -2 * self:Dot( normal ) + self
end

function Vector2:Angle( other )
	return math.atan( self.x , -self.y ) - math.atan( other.x , -other.y );
end

--
-- STATIC METHODS
--

Vector2.Zero = Vector2( 0, 0 )
Vector2.One = Vector2( 1, 1 )
Vector2.XPos = Vector2( 1, 0 )
Vector2.XNeg = Vector2( -1, 0 )
Vector2.YPos = Vector2( 0, 1 )
Vector2.YNeg = Vector2( 0, -1 )

function Vector2.Bezier3( v1, v2, v3, t )
	return Vector2( bezier3( v1.x, v2.x, v3.x, t ),
					bezier3( v1.y, v2.y, v3.y, t ) )
end

function Vector2.Bezier4( v1, v2, v3, v4, t )
	return Vector2( bezier4( v1.x, v2.x, v3.x, v4.x, t ),
					bezier4( v1.y, v2.y, v3.y, v4.y, t ) )
end
